'''
.. moduleauthor:: Rasmus Diederichsen

This module contains the base definition for subscriber functionality. The
:class:`~ikkuna.export.subscriber.Subscriber` class should be subclassed for adding new metrics.

'''
import abc
from collections import defaultdict
import ikkuna.visualization
from ikkuna.export.messages import MessageBundle, ModuleMessage, get_default_bus


class Subscription(object):
    '''Specification for a subscription that can span multiple kinds and a tag.

    Attributes
    ----------
    _tag    :   str
                Tag for filtering the processed messages
    _subscriber :   ikkuna.export.subscriber.Subscriber
                    The subscriber associated with the subscription
    counter :   dict(tuple(ikkuna.utils.NamedModule, str) or str, int)
                Number of times the Subscription was called for each pair of module and kind or meta
                data identifier. Since one :class:`ikkuna.export.subscriber.Subscriber` is
                associated with only one configuration of
                :class:`ikkuna.export.messages.MessageBundle`, this will enable proper subsampling
                of message streams.
    kinds   :   list(str)
                List of string identifiers for different message kinds. These are all the
                message kinds the subscriber wishes to receive
    _subsample  :   int
                    Factor for subsampling incoming messages. Only every ``subsample``-th
                    message will be processed.
    '''

    def __init__(self, subscriber, kinds, tag='default', subsample=1):
        '''
        Parameters
        ----------
        subscriber  :   ikkuna.export.subscriber.Subscriber
                        Object that wants to receive the messages
        kinds   :   list
                    List of message kinds to process
        tag :   str
                tag for filtering messages
        subsample   :   int
                        Number of messages to ignore before processing one. Note that this number if
                        applied to every kind, regardless of frequency. So if ``subsample = 10``,
                        every tenth ``weights`` message would be processed, but also only every
                        tenth ``epoch_finished`` message.
        tag :   str
                Optional tag for filtering messages.
        '''
        if not isinstance(kinds, list):
            raise ValueError(f'Expected list of kinds, got {type(kinds)}')
        self._tag        = tag
        self._subscriber = subscriber

        self._kinds      = kinds
        self._counter    = defaultdict(int)
        self._subsample  = subsample

    @property
    def counter(self):
        return dict(self._counter)

    @property
    def kinds(self):
        return self._kinds

    def _handle_message(self, message):
        '''Process a newly arrived message. Subclasses should override this method for any special
        treatment.

        Parameters
        ----------
        message    :   ikkuna.export.messages.Message
        '''
        self._subscriber.process_messages(message)

    def handle_message(self, message):
        '''Callback for receiving an incoming message.

        Parameters
        ----------
        message    :   ikkuna.export.messages.ModuleMessage
        '''
        if message.kind not in self.kinds or self._tag != message.tag:
            return

        if isinstance(message, ModuleMessage):
            key = (message.module, message.kind)
        else:
            key = message.kind
        if self._counter[key] % self._subsample == 0:
            self._handle_message(message)
        self._counter[key] += 1


class SynchronizedSubscription(Subscription):
    '''A subscription which buffers messages and publishes a set of messages, each of a different
    kind, when one round (a train step) is over. This is useful for receiving several kinds of
    messages in each train step and always have them be processed together.'''

    def __init__(self, subscriber, kinds, tag='default', subsample=1):
        super().__init__(subscriber, kinds, tag, subsample)
        self._current_global_step = None
        self._open_bundles        = {}

    def _new_round(self, round_idx):
        '''Start a new round of buffering, clearing the previous cache and resetting the record for
        which kinds were received in this round.

        Parameters
        ----------
        round_idx : int
                    Global train step of the new round

        Raises
        ------
        RuntimeError
            If not all desired kinds have been received for all identifiers yet in the current round
        '''
        for bundle in self._open_bundles.values():
            if not bundle.complete():
                raise RuntimeError(f'Bundle with id={bundle.key} not yet complete.')
        self._current_global_step = round_idx
        self._open_bundles = {}

    def _publish_complete(self):
        delete_these = []
        # any full? publish
        for module_and_kind, message_bundle in self._open_bundles.items():
            if message_bundle.complete():
                self._subscriber.process_messages(message_bundle)
                delete_these.append(module_and_kind)

        # purge published data
        for module_and_kind in delete_these:
            del self._open_bundles[module_and_kind]

    def _handle_message(self, message):
        '''Start a new round if a new sequence number is seen.'''

        # if we get a new sequence number, a new train step must have begun
        if self._current_global_step is None or self._current_global_step != message.global_step:
            self._new_round(message.global_step)

        # module not seen -> init data
        key = message.key
        if key not in self._open_bundles:
            self._open_bundles[key] = MessageBundle(self.kinds)
        self._open_bundles[key].add_message(message)

        self._publish_complete()


class Subscriber(abc.ABC):
    '''Base class for receiving and processing activations, gradients and other stuff into
    insightful metrics.'''

    def __init__(self, subscriptions, message_bus):
        '''
        Parameters
        ----------
        subscriptions    :   list(Subscription)
        '''
        self._subscriptions    = {kind: subscription
                                  for subscription in subscriptions for kind in subscription.kinds}
        self._msg_bus          = message_bus
        message_bus.register_subscriber(self)
        self._published_topics = defaultdict(list)

    def _add_publication(self, topic, type='DATA'):
        '''
        Parameters
        ----------
        topic   :   str
                    Topic name
        type    :   str
                    ``DATA`` or ``META``
        '''
        if type not in ('DATA', 'META'):
            raise ValueError(f'Unknown message type "{type}"')

        self._published_topics[type].append(topic)
        if type == 'DATA':
            self._msg_bus.register_data_topic(topic)
        else:
            self._msg_bus.register_meta_topic(topic)

    def __del__(self):
        '''If for whatever reason a subscriber ceases to exist before the interpreter ends, delete
        the registered topics'''
        for topic in self._published_topics['DATA']:
            self._msg_bus.deregister_data_topic(topic)
        for topic in self._published_topics['META']:
            self._msg_bus.deregister_meta_topic(topic)

    @property
    def publications(self):
        return {k: v for k, v in self._published_topics.items()}

    @property
    def kinds(self):
        return list(set(self.subscriptions.keys()))

    @property
    def subscriptions(self):
        return self._subscriptions

    @property
    def message_bus(self):
        return self._msg_bus

    @abc.abstractmethod
    def compute(self, message_or_bundle):
        '''This is where the magic happens. Subclasses should override this method so that they can
        compute their metric upon reception of their desired messages or do whatever else they want.
        If interested in plotting, they should then use their
        :attr:`~ikkuna.export.subscriber.PlotSubscriber.backend` property to plot the metric (if
        they display line plots) and their ``message_bus`` to publish a new message with the metric.

        Parameters
        ----------
        message_or_bundle : ikkuna.export.messages.Message or ikkuna.export.messages.MessageBundle
                            If the subscriber uses a :class:`SynchronizedSubscription`, a bundle is
                            received in each call, otherwise just a
                            :class:`~ikkuna.export.messages.Message`.

                            Messages can either be :class:`~ikkuna.export.messages.NetworkMessage`
                            if the Subscriber is not interested in actual training artifacts, or
                            :class:`~ikkuna.export.messages.ModuleMessage`
        '''
        pass

    def receive_message(self, message):
        '''Process a single message received from an :class:`~ikkuna.export.messages.MessageBus`.'''

        if message.kind in self._subscriptions:
            self._subscriptions[message.kind].handle_message(message)

    def process_messages(self, message_or_bundle):
        '''Callback for processing a single :class:`~ikkuna.export.messages.Message` or a
        :class:`~ikkuna.export.messages.MessageBundle` object with
        :class:`~ikkuna.export.messages.NetworkMessage`\ s or
        :class:`~ikkuna.export.messages.ModuleMessage`\ s in it.

        Parameters
        ----------
        message_or_bundle : ikkuna.export.messages.Message or ikkuna.export.messages.MessageBundle
                            The exact nature of this package is determined by the
                            :class:`ikkuna.export.subscriber.Subscription` attached to this
                            :class:`ikkuna.export.subscriber.Subscriber`.

        Raises
        ------
        ValueError
            If the received :class:`~ikkuna.export.messages.MessageBundle` object is not
            :meth:`~ikkuna.export.messages.MessageBundle.complete()`
        '''
        if (isinstance(message_or_bundle, ikkuna.export.messages.MessageBundle)
            and not message_or_bundle.complete()):
            raise ValueError(f'Data received for "{message_or_bundle._module}" is not complete.')

        self.compute(message_or_bundle)


class PlotSubscriber(Subscriber):
    '''Base class for subscribers that output scalar or histogram values per time and module

    Attributes
    ----------
    _backend    :   ikkuna.visualization.Backend
                    Plotting backend
    '''

    def __init__(self, subscriptions, message_bus, plot_config, backend='tb'):
        '''
        Parameters
        ----------
        ylims   :   tuple(int, int)
                    Optional Y-axis limits
        plot_config :   dict
                        Configuration parameters for plotting. Relevant keys are ``title``,
                        ``xlabel``, ``ylabel`` and ``ylims``. Which of them are actually used
                        depends on the :class:`~ikkuna.visualization.Backend`
        '''
        super().__init__(subscriptions, message_bus)

        self._backend = ikkuna.visualization.get_backend(backend, plot_config)

    @property
    def backend(self):
        '''ikkuna.visualization.Backend: The backend to use for plotting'''
        return self._backend

    @abc.abstractmethod
    def compute(self, message_or_bundle):
        pass


class CallbackSubscriber(Subscriber):
    '''Subscriber class for subscribing to :class:`~ikkuna.export.messages.ModuleMessage`\ s and
    running a callback with them.'''

    def __init__(self, kinds, callback, message_bus=get_default_bus(), tag='default', subsample=1):
        '''
        Parameters
        ----------
        subscription    :   Subscription
        message_bus :   ikkuna.export.messages.MessageBus
        callback    :   function
                        A function that accepts as many parameters as the ``Subscription`` delivers
                        messages at once.
        '''
        subscription = SynchronizedSubscription(self, kinds, tag, subsample)
        super().__init__([subscription], message_bus)
        self._callback = callback

    def compute(self, message_or_bundle):
        kinds = self._subscriptions[0].kinds
        args  = (message_or_bundle.data[kind] for kind in kinds)

        id_ = message_or_bundle.key
        self._callback(*args, message_or_bundle.global_step, message_or_bundle.train_step,
                       message_or_bundle.epoch, id_)



'''
MIT License

Copyright (c) 2017 liukuang

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.
'''

'''I tried torchvision's model code initially, but that one does not deal with image sizes smaller
than imagnet. An alternative implementation is `here
<https://github.com/kuangliu/pytorch-cifar/blob/master/models/resnet.py>`_. I'm not sure how faithful
the reproduction is, but ultimately it shouldn't matter.'''
import torch
import torch.nn as nn
import torch.nn.functional as F


class BasicBlock(nn.Module):
    expansion = 1

    def __init__(self, in_planes, planes, stride=1):
        super(BasicBlock, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.bn2(self.conv2(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class Bottleneck(nn.Module):
    expansion = 4

    def __init__(self, in_planes, planes, stride=1):
        super(Bottleneck, self).__init__()
        self.conv1 = nn.Conv2d(in_planes, planes, kernel_size=1, bias=False)
        self.bn1 = nn.BatchNorm2d(planes)
        self.conv2 = nn.Conv2d(planes, planes, kernel_size=3, stride=stride, padding=1, bias=False)
        self.bn2 = nn.BatchNorm2d(planes)
        self.conv3 = nn.Conv2d(planes, self.expansion*planes, kernel_size=1, bias=False)
        self.bn3 = nn.BatchNorm2d(self.expansion*planes)

        self.shortcut = nn.Sequential()
        if stride != 1 or in_planes != self.expansion*planes:
            self.shortcut = nn.Sequential(
                nn.Conv2d(in_planes, self.expansion*planes, kernel_size=1, stride=stride, bias=False),
                nn.BatchNorm2d(self.expansion*planes)
            )

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = F.relu(self.bn2(self.conv2(out)))
        out = self.bn3(self.conv3(out))
        out += self.shortcut(x)
        out = F.relu(out)
        return out


class ResNet(nn.Module):
    def __init__(self, BlockType, num_blocks, num_classes=10, exporter=None):
        super(ResNet, self).__init__()
        self.in_planes = 64

        self.conv1 = nn.Conv2d(3, 64, kernel_size=3, stride=1, padding=1, bias=False)
        self.bn1 = nn.BatchNorm2d(64)
        self.layer1 = self._make_layer(BlockType, 64, num_blocks[0], stride=1)
        self.layer2 = self._make_layer(BlockType, 128, num_blocks[1], stride=2)
        self.layer3 = self._make_layer(BlockType, 256, num_blocks[2], stride=2)
        self.layer4 = self._make_layer(BlockType, 512, num_blocks[3], stride=2)
        self.linear = nn.Linear(512*BlockType.expansion, num_classes)

        if exporter:
            exporter.set_model(self)
            exporter.add_modules(self)

    def _make_layer(self, BlockType, planes, num_blocks, stride):
        strides = [stride] + [1]*(num_blocks-1)
        layers = []
        for stride in strides:
            layers.append(BlockType(self.in_planes, planes, stride))
            self.in_planes = planes * BlockType.expansion
        return nn.Sequential(*layers)

    def forward(self, x):
        out = F.relu(self.bn1(self.conv1(x)))
        out = self.layer1(out)
        out = self.layer2(out)
        out = self.layer3(out)
        out = self.layer4(out)
        out = F.avg_pool2d(out, 4)
        out = out.view(out.size(0), -1)
        out = self.linear(out)
        return out


def resnet18(**kwargs):
    '''Constructs a ResNet-18 model.

    Parameters
    ----------
    kwargs  :   dict
                ResNet parameters
    '''
    return ResNet(BasicBlock, [2, 2, 2, 2], **kwargs)


def resnet34(pretrained=False, **kwargs):
    '''Constructs a ResNet-34 model.

    Parameters
    ----------
    kwargs  :   dict
                ResNet parameters
    '''
    return ResNet(BasicBlock, [3, 4, 6, 3], **kwargs)


def resnet50(pretrained=False, **kwargs):
    '''Constructs a ResNet-50 model.

    Parameters
    ----------
    kwargs  :   dict
                ResNet parameters
    '''
    return ResNet(Bottleneck, [3, 4, 6, 3], **kwargs)


def resnet101(pretrained=False, **kwargs):
    '''Constructs a ResNet-101 model.

    Parameters
    ----------
    kwargs  :   dict
                ResNet parameters
    '''
    return ResNet(Bottleneck, [3, 4, 23, 3], **kwargs)


def resnet152(**kwargs):
    '''Constructs a ResNet-152 model.

    Parameters
    ----------
    kwargs  :   dict
                ResNet parameters
    '''
    return ResNet(Bottleneck, [3, 8, 36, 3], **kwargs)



from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class LossSubscriber(PlotSubscriber):
    '''A Subscriber that plots the training loss at every time step.'''

    def __init__(self, message_bus=get_default_bus(), tag='default', subsample=1, ylims=None,
                 backend='tb'):

        title        = 'Loss'
        ylabel       = 'loss'
        xlabel       = 'Train step'
        subscription = Subscription(self, ['loss'], tag=tag, subsample=subsample)
        super().__init__([subscription], message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)

    def compute(self, message):
        data = message.data
        self._backend.add_data('loss', data, message.global_step)



import torch
from torch.utils.data import DataLoader

from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus


class TestAccuracySubscriber(PlotSubscriber):
    '''
    Subscriber which can compute the accuracy on a given data set (generally the test/validation
    set).

    Attributes
    ----------
    _dataset_meta    :   train.DatasetMeta
                         Dataset to compute accuracy over
    _data_loader    :   torch.utils.data.DataLoader
                        Loader on the dataset
    _frequency  :   int
                    Number of batches to ignore before computing accuracy once
    _forward_fn :   function
                    Bound method on the model to push data through and get predictions
    '''

    def __init__(self, dataset_meta, forward_fn, batch_size, message_bus=get_default_bus(),
                 frequency=100, ylims=None, tag='default', subsample=1, backend='tb'):
        '''
        Parameters
        ----------
        dataset_meta    :   train.DatasetMeta
                            Test dataset
        forward_fn  :   function
                        Bound version of :meth:`torch.nn.Module.forward()`. This could be obtained
                        from the :attr:`train.Trainer.model` property.
        batch_size  :   int
                        Batch size to use for pushing the test data through the model. This doesn't
                        have to be the training batch size, but must be selected so that the forward
                        pass fits into memory.
        frequency   :   int
                        Inverse of the frequency with which to compute the accuracy
        '''
        kinds  = ['batch_finished']
        title  = f'test_accuracy'
        ylabel = 'Accuracy'
        xlabel = 'Train step'
        subscription = Subscription(self, kinds, tag=tag, subsample=subsample)
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel,
                          'redraw_interval': 1},
                         backend=backend)

        self._dataset_meta = dataset_meta
        self._data_loader  = DataLoader(dataset_meta.dataset, batch_size=batch_size, shuffle=False,
                                        pin_memory=True)
        self._frequency    = frequency
        self._forward_fn   = forward_fn

        self._add_publication('test_accuracy', type='META')

    def compute(self, message):
        '''Compute accuracy over the entire test set.

        A :class:`~ikkuna.export.messages.NetworkMessage` with the identifier
        ``test_accuracy`` is published.
        '''
        if self.subscriptions['batch_finished'].counter['batch_finished'] % self._frequency == 0:
            n_batches = 0
            accuracy  = 0
            loader    = iter(self._data_loader)
            try:
                while True:
                    X, labels   = next(loader)
                    if torch.cuda.is_available():
                        X = X.cuda()
                    outputs     = self._forward_fn(X, should_train=False, tag='testing')
                    predictions = outputs.argmax(1)
                    n_correct   = (predictions.cpu() == labels).sum().item()
                    accuracy   += n_correct / X.shape[0]
                    n_batches  += 1
            except StopIteration:
                accuracy /= n_batches
                self._backend.add_data('test_accuracy', accuracy, message.global_step)

                kind = 'test_accuracy'
                self.message_bus.publish_network_message(message.global_step,
                                                         message.train_step,
                                                         message.epoch, kind,
                                                         data=accuracy)



'''
.. moduleauthor Rasmus Diederichsen

This module defines :class:`ikkuna.models.AlexNetMini`, a reduced version of AlexNet.
Adapted from
https://github.com/sukilau/Ziff-deep-learning/blob/master/3-CIFAR10-lrate/CIFAR10-lrate.ipynb
'''
import torch


class AlexNetMini(torch.nn.Module):     # Pytorch Sphinx-doc is buggy here, so use full path
    '''Reduced AlexNet (basically just a few conv layers with relu and
    max-pooling) which attempts to adapt to arbitrary input sizes, provided they are large enough to
    survive the strides and conv cutoffs.

    Attributes
    ---------
    features    :   torch.nn.Module
                    Convolutional module, extracting features from the input
    classifier  :   torch.nn.Module
                    Classifier with relu and dropout
    H_out   :   int
                Output height of the classifier
    W_out   :   int
                Output width of the classifier
    '''
    def __init__(self, input_shape, num_classes=1000, exporter=None):
        '''
        Parameters
        ----------
        input_shape :   tuple
                        H, W C (beware of channel order, is different from what you call the model
                        with)
        '''
        super(AlexNetMini, self).__init__()

        # if channel dim not present, add 1
        if len(input_shape) == 2:
            input_shape.append(1)
        H, W, C = input_shape

        self.features = torch.nn.Sequential(
            torch.nn.Conv2d(C, 4, kernel_size=3, padding=0),
            torch.nn.ReLU(inplace=False),
            torch.nn.Conv2d(4, 8, kernel_size=3, padding=0),
            torch.nn.ReLU(inplace=False),
            torch.nn.MaxPool2d(kernel_size=2, stride=2),
        )
        # for each conv or padding layer, apply the output size formula ((length - filter_size +
        # 2*padding)/ stride) + 1
        H_conv1    = (H - 3 + 2 * 0) // 1 + 1
        H_conv2    = (H_conv1 - 3 + 2 * 0) // 1 + 1
        self.H_out = (H_conv2 - 2 + 2 * 0) // 2 + 1

        W_conv1    = (W - 3 + 2 * 0) // 1 + 1
        W_conv2    = (W_conv1 - 3 + 2 * 0) // 1 + 1
        self.W_out = (W_conv2 - 2 + 2 * 0) // 2 + 1

        self.classifier = torch.nn.Sequential(
            torch.nn.Dropout(0.25),
            torch.nn.Linear(8 * self.H_out * self.W_out, 16),
            torch.nn.ReLU(inplace=False),
            torch.nn.Dropout(0.5),
            torch.nn.Linear(16, num_classes),
            # torch.nn.Softmax()
        )

        if exporter:
            exporter.set_model(self)
            exporter.add_modules(self)

    def forward(self, x):
        x = self.features(x)
        x = x.view(x.size(0), 8 * self.H_out * self.W_out)
        x = self.classifier(x)
        return x




from ikkuna.export.subscriber import PlotSubscriber, Subscription
from ikkuna.export.messages import get_default_bus

try:
    from hessian_eigenthings import compute_hessian_eigenthings
except ImportError:
    print('You need to install `https://github.com/noahgolmant/pytorch-hessian-eigenthings/`')
    import sys
    sys.exit(1)


class HessianEigenSubscriber(PlotSubscriber):
    '''A subscriber to compute the top-k eigenvalues of the hessian of the loss w.r.t. the weights.
    This is done by using defalted power iteration from Noah Golmant's ``hessian_eigenthings``
    module, which has to be installed. This operation is _very_ expensive, since it involves
    differentiating twice over a subset of the training data. Since the weights must be fixed,
    gradients from training cannot be reused.

    .. warning::
        In my experiments, the power iteration from ``hessian_eigenthings`` does not really work;
        the estimates do not converge to a single value when increasing the number of steps, but
        rather are all over the place. Investigation needed.
    '''

    def __init__(self, forward_fn, loss_fn, data_loader, batch_size, frequency=1, num_eig=1,
                 power_steps=20, tag='default', message_bus=get_default_bus(), ylims=None,
                 backend='tb'):
        '''
        Parameters
        ----------
        forward_fn  :   function
                        Function to obtain predictions. You probably want to pass the model's
                        ``forward()`` routine here
        loss_fn     :   torch.nn.Module
                        Loss function (such as :class:`torch.nn.CrossEntropyLoss`)
        data_loader :   torch.utils.data.DataLoader
                        Loader for the dataset to compute gradients over
        batch_finished  :   int
                            Number of samples to compute gradients for in one step of power iteration
                            More should lead to a better estimate
        frequency   :   int
                        How often to compute the eigenvalues (after every nth batch)
        num_eig :   int
                    Number of top eigenvalues to compute
        power_iter_steps    :   int
                                Number of steps in the power iteration for computing a singular value.
                                The total number of batches read is then ``power_iter_steps * num_eig``
                                More steps should lead to a better estimate.
        '''
        title  = f'Top hessian Eigenvalues'
        ylabel = 'tbd'
        xlabel = 'Train step'
        subscription = Subscription(self, ['batch_finished', 'activations'], tag=tag)
        super().__init__([subscription],
                         message_bus,
                         {'title': title,
                          'ylabel': ylabel,
                          'ylims': ylims,
                          'xlabel': xlabel},
                         backend=backend)

        self._forward_fn = forward_fn
        self._loss_fn = loss_fn
        self._power_steps = power_steps
        self._dataloader = data_loader
        self._parameters = set()
        self._num_eig = num_eig
        self._frequency = frequency

    def compute(self, message):
        if message.kind == 'activations':
            module = message.key.module
            for p in module.parameters():
                self._parameters.add(p)
        else:
            if self.subscriptions['batch_finished'].counter['batch_finished'] % self._frequency == 0:
                evals, evecs = compute_hessian_eigenthings(self._forward_fn,
                                                           self._parameters,
                                                           self._dataloader,
                                                           self._loss_fn,
                                                           power_iter_steps=self._power_steps,
                                                           num_eigenthings=self._num_eig)
                for i, val in enumerate(sorted(evals)):
                    self.backend.add_data(f'Eigenvalue {i}', evals[i], message.global_step)



import torch


class FullyConnectedModel(torch.nn.Module):

    def __init__(self, input_shape, num_classes=1000, exporter=None):
        super(FullyConnectedModel, self).__init__()

        # if channel dim not present, add 1
        if len(input_shape) == 2:
            input_shape = list(input_shape) + [1]
        H, W, C = input_shape

        self.features = torch.nn.Sequential(
            torch.nn.Linear(H * W * C, 4000),
            torch.nn.ReLU(inplace=False),
            torch.nn.Linear(4000, 4000),
            torch.nn.ReLU(inplace=False),
            torch.nn.Linear(4000, 4000),
            torch.nn.ReLU(inplace=False)
        )
        self.classifier = torch.nn.Sequential(
            torch.nn.Linear(4000, 1000),
            torch.nn.ReLU(inplace=False),
            torch.nn.Linear(1000, num_classes),
        )

        if exporter:
            exporter.set_model(self)
            exporter.add_modules(self)

    def forward(self, x):
        B, C, H, W = x.shape
        x = x.view(B, C * H * W)
        x = self.features(x)
        x = self.classifier(x)
        return x



#!/usr/bin/env python

from distutils.core import setup
import setuptools
import os
import ikkuna

with open('requirements.txt', mode='r') as f:
    requirements = f.read()
    required_pkgs, required_repos = requirements.split('# git repos')
    required_pkgs = required_pkgs.split()
    required_repos = required_repos.split()

with open('README.md') as f:
    readme = f.read()

setup(name='ikkuna',
      version=ikkuna.__version__,
      description='Ikkuna Neural Network Monitor',
      long_description=readme,
      author='Rasmus Diederichsen',
      author_email='rasmus@peltarion.com',
      url='https://peltarion.github.io/ai_ikkuna/',
      classifiers=['Development Status :: 3 - Alpha',
                   'Intended Audience :: Science/Research',
                   'Topic :: Scientific/Engineering :: Visualization',
                   'Topic :: Scientific/Engineering :: Artificial Intelligence',
                   'Programming Language :: Python :: 3.6',
                   'License :: OSI Approved :: MIT License',
                   'Intended Audience :: Developers',
                   ],
      keywords='deep-learning pytorch neural-networks machine-learning'.split(),
      packages=setuptools.find_packages('.', include=['ikkuna.*']),
      install_requires=required_pkgs,
      dependency_links=required_repos,
      entry_points={
          'ikkuna.export.subscriber': [
              'HistogramSubscriber = ikkuna.export.subscriber.histogram:HistogramSubscriber',
              'RatioSubscriber = ikkuna.export.subscriber.ratio:RatioSubscriber',
              'SpectralNormSubscriber = ikkuna.export.subscriber.spectral_norm:SpectralNormSubscriber',
              'TestAccuracySubscriber = ikkuna.export.subscriber.test_accuracy:TestAccuracySubscriber',
              'TrainAccuracySubscriber = ikkuna.export.subscriber.train_accuracy:TrainAccuracySubscriber',
              'VarianceSubscriber = ikkuna.export.subscriber.variance:VarianceSubscriber',
              'NormSubscriber = ikkuna.export.subscriber.norm:NormSubscriber',
              'MeanSubscriber = ikkuna.export.subscriber.mean:MeanSubscriber',
              'HessianEigenSubscriber = ikkuna.export.subscriber.hessian_eig:HessianEigenSubscriber',
              'MessageMeanSubscriber = ikkuna.export.subscriber.message_mean:MessageMeanSubscriber',
              'LossSubscriber = ikkuna.export.subscriber.loss:LossSubscriber',
              'SVCCASubscriber = ikkuna.export.subscriber.svcca:SVCCASubscriber',
              'BatchedSVCCASubscriber = ikkuna.export.subscriber.svcca:BatchedSVCCASubscriber',
          ]
      },
      zip_safe=False,   # don't install egg, but source
      )
